/*
    Copyright (c) 2005-2024 Leisenfels GmbH. All rights reserved.
    Use is subject to license terms.
*/

package com.lf.javainfo.gui;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.*;
import java.net.URL;
import java.text.Collator;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.*;


/**
 * Small GUI/CLI application to let the user send Java system properties to a servlet.
 * <p/>
 * This class was inspired by http://lopica.sourceforge.net/os.html providing a list of possible
 * values for the <code>os.name</code>, <code>os.version</code>, and <code>os.arch</code>
 * properties found on different platforms. This information can be gathered by a web database
 * and may be a basis for values used with Web Start applications where platform specific
 * native libs are needed.
 * <p/>
 * By default this application opens a simple dialog with buttons to send data to the
 * web servlet or to save the data into a file. If started with the "-headless" parameter
 * then the output is printed on the command line (stdout). This is necessary for VMs
 * like those on iSeries machines (aka AS/400) which lack GUI support. Most operating
 * systems provide a way to save this output into a file which can then be copied to a
 * special web form.
 * <p/>
 * In order to prevent web robots from sending arbitrary content to the servlet, the XML
 * file contains a checksum generated from certain parts of the XML data. This way we can
 * make sure that only data portions generated by this executable is sent to the servlet.
 * <p/>
 * The current JNLP specification doesn't list the values for the os or arch attributes.
 * Instead it states that all values are valid as long as they match the values returned
 * by the system properties <code>os.name</code> or <code>os.arch</code>. Values that only
 * match the beginning of <code>os.name</code> or <code>os.arch</code> are also valid
 * (e.g. if you specify Windows and the <code>os.name</code> is Windows 98, for example,
 * everything works out fine).
 * <p/>
 * Remember that this class must be compatible with Java 1.4 (!) in order to support even
 * older Java VMs and is therefore very limited in GUI design and the set of Java features.
 * For this reason this class includes several methods from other classes of this package
 * since it must be runnable standalone.
 * <p/>
 * The API documentation of Java 1.1.8 lists the following properties to be available:
 * <p/>
 * <pre>
 *     java.version         Java version number
 *     java.vendor          Java vendor-specific string
 *     java.vendor.url      Java vendor URL
 *     java.home            Java installation directory
 *     java.class.version   Java class format version number
 *     java.class.path      Java class path
 *     os.name              Operating system name
 *     os.arch              Operating system architecture
 *     os.version           Operating system version
 *     file.separator       File separator ("/" on UNIX)
 *     path.separator       Path separator (":" on UNIX)
 *     line.separator       Line separator ("\n" on UNIX)
 *     user.name            User's account name
 *     user.home            User's home directory
 *     user.dir             User's current working directory
 * </pre>
 * <p/>
 *
 * @author Axel Schwolow
 * @created 2016-01-01
 * @since 1.6
 */
public class JavaInfo extends Applet {


    /**
     * The icon to be displayed by the window.
     */
    public final static String ICONPATH = "/com/lf/javainfo/resource/cup_32.png";
    /**
     * The deploy properties file.
     */
    public final static String PROPSPATH = "/com/lf/javainfo/resource/deploy.properties";
    /**
     * The default name.
     */
    public final static String JAVAINFO_APP_NAME = "JavaInfo";
    /**
     * The URL to upload Java properties.
     */
    public static String SERVLETURL = "n/a";

    /**
     * Constructor method for i18n purposes only.
     * <p/>
     * Can only be used if beans design time is set (e.g. by external tools like the
     * <code>I18NExtractor</code>).
     * <p/>
     *
     * @throws InstantiationException Error indication
     * @since 1.6
     */
    public JavaInfo() throws InstantiationException {

        if (!java.beans.Beans.isDesignTime()) {
            throw new InstantiationException("Do not use this constructor!");
        }
    }

    /**
     * Implements the headless mode of this application.
     * <p/>
     *
     * @since 1.6
     */
    private static void processCLI() {

        try {
            Properties props = new Properties();  // Avoid using JavaInfoSettings (Java 1.4)!
            try {
                props.load(JavaInfo.class.getResourceAsStream(PROPSPATH));
                SERVLETURL = props.getProperty("SERVLETURL");
            } catch (Exception ignored) {
            }

            // Print the system properties first
            byte[] data = JavaInfoUtils.generateData(JAVAINFO_APP_NAME);
            System.out.write(data);

            // Main menu with button functionality
            System.out.println("\n1 - " + getLocalizedText("JavaInfo.submit"));
            System.out.println("2 - " + getLocalizedText("JavaInfo.saveas"));
            System.out.println("3 - " + getLocalizedText("JavaInfo.exit") + '\n');

            String input = JavaInfo.getInput(getLocalizedText("JavaInfo.choose") + ' ', new String[]{"1", "2", "3"});
            if (input.equals("1")) {

                // Try to submit the properties to web service
                String address = SERVLETURL + "?locale=" + JavaInfoUtils.LF_LOCALE;
                byte[] result = JavaInfoUtils.callWebServiceByPOST(new URL(address), data);

                // Status OK? If not, the service possibly answered with an error message.
                String status = JavaInfoUtils.checkResponse(result);
                if (status != null && status.equals("OK")) {
                    System.out.println('\n' + getLocalizedText("JavaInfo.ok_submit") + '\n');
                } else {
                    if (status != null) System.out.println('\n' + status + '\n');  // Print directly
                    else {
                        System.out.println('\n' + getLocalizedText("JavaInfo.error_submit") + '\n');
                    }
                }
            } else if (input.equals("2")) {

                // Save data to file. Ask for file name and check if the file already exists.
                System.out.println("");
                String filename = null;
                while (filename == null) {
                    filename = JavaInfo.getInput(getLocalizedText("JavaInfo.filename") + ' ',
                            new String[0]);
                    if (filename.trim().length() == 0) filename = null;
                }

                File target = new File(filename);
                if (!target.isAbsolute()) {

                    // Expand relative file name like 'props.xml' with working directory
                    String workdir = System.getProperties().getProperty("user.dir");
                    target = new File(workdir + File.separator + filename);
                }
                if (target.exists()) {
                    String text = getLocalizedText("JavaInfo.askoverwrite");
                    text = new String(JavaInfoUtils.replace(text.toCharArray(), "%FILE%", target.toString()));
                    input = JavaInfo.getInput('\n' + text + ' ',
                            new String[]{getLocalizedText("JavaInfo.askyes"),
                                    getLocalizedText("JavaInfo.askno")});
                    if (input.toLowerCase().equals(getLocalizedText("JavaInfo.askno"))) {
                        System.out.println('\n' + getLocalizedText("JavaInfo.notoverwritten"));
                        return;  // Exit
                    }
                }

                // Ok, let's store the data into the specified file
                try {
                    FileOutputStream fstream = new FileOutputStream(target);
                    fstream.write(data);
                    fstream.flush();
                    fstream.close();
                    System.out.println('\n' + getLocalizedText("JavaInfo.ok_write") + '\n');
                } catch (Exception e) {
                    System.out.println('\n' + getLocalizedText("JavaInfo.error_write") + '\n');
                }
            }
        } catch (Exception e) {
        }
    }

    /**
     * Gets input from the command line.
     * <p/>
     * If the blank string is allowed, then add it to the <code>expected</code> array.
     * <p/>
     *
     * @param output   The label to print
     * @param expected List of possible values
     * @return One of the expected values
     */
    private static String getInput(String output, String[] expected) {

        String input;

        while (true) {
            System.out.print(output);
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            try {
                input = br.readLine();
                if (expected != null && expected.length >= 1) {
                    for (int i = 0; i < expected.length; i++) {
                        if (input.equals(expected[i])) return input;
                    }
                } else return input;  // Accept anything, if expected is not set
            } catch (Exception e) {
            }
        }
    }

    /**
     * Implements the GUI-based mode of this application.
     * <p/>
     *
     * @since 1.6
     */
    private static void processGUI() {

        String name = JavaInfo.class.getName();  // getSimpleName() is still being invented ;-)
        name = name.substring(name.lastIndexOf('.') + 1);
        String title = name;
        Properties props = new Properties();  // Avoid using JavaInfoSettings (Java 1.4)!
        try {
            props.load(JavaInfo.class.getResourceAsStream(PROPSPATH));
            title = props.getProperty("APPLICATION") + ' ' + props.getProperty("VERSION");
            SERVLETURL = props.getProperty("SERVLETURL");
        } catch (Exception ignored) {
        }

        // Construct the AWT window with the few components
        final Frame frame = new Frame(title);
        frame.setLayout(new BorderLayout(5, 5));
        Panel dummy = new Panel();
        dummy.setLayout(new BorderLayout(0, 0));
        frame.add(dummy, BorderLayout.NORTH);
        dummy = new Panel();
        dummy.setLayout(new BorderLayout(0, 0));
        frame.add(dummy, BorderLayout.WEST);
        dummy = new Panel();
        dummy.setLayout(new BorderLayout(0, 0));
        frame.add(dummy, BorderLayout.EAST);

        Toolkit toolkit = frame.getToolkit();
        final Dimension dims = toolkit.getScreenSize();

        URL url = JavaInfo.class.getResource(JavaInfo.ICONPATH);
        Image image = Toolkit.getDefaultToolkit().getImage(url);
        frame.setIconImage(image);

        // Add the components manually. Do not under Mac OS, scroller shows up automatically.
        TextArea textarea = new TextArea();
        if (System.getProperty("os.name", "").equals("Mac OS X")) {
            textarea.setEditable(false);
            frame.add(textarea, BorderLayout.CENTER);
        } else {
            ScrollPane scroller = new ScrollPane();
            textarea.setEditable(false);
            scroller.add(textarea, BorderLayout.CENTER);
            frame.add(scroller, BorderLayout.CENTER);
        }

        Panel bottom = new Panel(new BorderLayout(0, 0));
        Panel bottomright = new Panel(new FlowLayout(FlowLayout.RIGHT, 10, 10));
        final Button buttonSubmit = new Button(getLocalizedText("JavaInfo.submit"));
        bottomright.add(buttonSubmit);
        final Button buttonSave = new Button(getLocalizedText("JavaInfo.saveas"));
        bottomright.add(buttonSave);
        final Button buttonExit = new Button(getLocalizedText("JavaInfo.exit"));
        bottomright.add(buttonExit);
        buttonExit.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                System.exit(0);
            }
        });
        bottom.add(bottomright, BorderLayout.EAST);

        Panel bottomleft = new Panel(new FlowLayout(FlowLayout.LEFT, 10, 10));
        Label copylabel = new Label("  " + JavaInfoUtils.COPYRIGHT);  // HTML is not possible here
        final Label msglabel = new Label("                                                  ");
        bottomleft.add(msglabel);
        bottom.add(copylabel, BorderLayout.NORTH);
        bottom.add(bottomleft, BorderLayout.WEST);
        frame.add(bottom, BorderLayout.SOUTH);

        frame.pack();  // Necessary for Linux

        // Place the window in the center of the screen
        int defaultWidth = 580;
        int defaultHeight = 420;
        int defaultX = (int) ((dims.width - defaultWidth) / 2.0);
        int defaultY = (int) ((dims.height - defaultHeight) / 2.0) - 10;
        frame.setSize(defaultWidth, defaultHeight);
        frame.setLocation(defaultX, defaultY);

        WindowAdapter adapter = new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                System.exit(0);
            }
        };
        frame.addWindowListener(adapter);

        // Add the normalized system properties finally
        final byte[] data = JavaInfoUtils.generateData(JAVAINFO_APP_NAME);
        try {
            textarea.setText(new String(data, JavaInfoUtils.ENCODING));
        } catch (Exception e) {
        }

        buttonSave.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {

                FileDialog dialog = new FileDialog(frame, getLocalizedText("JavaInfo.saveas"),
                        FileDialog.SAVE);
                // Do not adjust the user dir here, looks strange on Mac OS X.
                // Normally, the file chooser automagically navigates to the user's home.
                dialog.setFile("props.xml");
                dialog.setVisible(true);
                if (dialog.getFile() == null) return;  // Cancelled

                try {
                    String sep = System.getProperties().getProperty("file.separator");
                    if (sep == null) sep = "/";
                    File target = new File(dialog.getDirectory() + sep + dialog.getFile());
                    FileOutputStream fstream = new FileOutputStream(target);
                    fstream.write(data);
                    fstream.flush();
                    fstream.close();
                    JavaInfo.setMessage(frame, msglabel, getLocalizedText("JavaInfo.ok_write"),
                            true, 10000);
                } catch (Exception e) {
                    JavaInfo.setMessage(frame, msglabel, getLocalizedText("JavaInfo.error_write"),
                            false, 10000);
                }
            }
        });

        buttonSubmit.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {

                // Don't freeze the EDT, submission may take some time or may block!
                final Thread thread = new Thread(new Runnable() {
                    public void run() {

                        buttonSubmit.setEnabled(false);
                        buttonSave.setEnabled(false);
                        String status = null;

                        try {
                            URL url = new URL(SERVLETURL + "?locale=" + JavaInfoUtils.LF_LOCALE);
                            //System.out.println("url=" + url);
                            byte[] result = JavaInfoUtils.callWebServiceByPOST(url, data);

                            // Status OK? If not, the service possibly returned an error message
                            status = JavaInfoUtils.checkResponse(result);
                        } catch (Exception e) {
                            if (e instanceof InterruptedException) {
                                status = getLocalizedText("JavaInfo.timeout");
                            }
                        } finally {
                            buttonSubmit.setEnabled(true);
                            buttonSave.setEnabled(true);

                            if (status != null && status.equals("OK")) {
                                String text = getLocalizedText("JavaInfo.ok_submit");
                                JavaInfo.setMessage(frame, msglabel, text, true, 10000);
                            } else {
                                if (status != null) {
                                    JavaInfo.setMessage(frame, msglabel, status, false, 10000);
                                } else {
                                    status = getLocalizedText("JavaInfo.error_submit");
                                    JavaInfo.setMessage(frame, msglabel, status, false, 10000);
                                }
                            }
                        }
                    }
                }, "[javainfo] Updater");
                thread.start();

                // And another thread to monitor the first thread
                new Thread(new Runnable() {
                    public void run() {
                        try {
                            thread.join(10000);  // Wait max. 10 seconds, then interrupt it
                        } catch (Exception e) {
                        }
                        try {
                            if (thread.isAlive()) {
                                thread.interrupt();
                            }
                        } catch (Exception e) {
                        }
                    }
                }, "[javainfo] Update Monitor").start();
            }
        });

        frame.setVisible(true);
    }

    /**
     * Prints a message for the user.
     * <p/>
     *
     * @param frame   The frame
     * @param label   The component
     * @param text    The text
     * @param success Successful?
     * @param millis  Shown for n milliseconds, nonstop if -1
     * @since 1.6
     */
    private static void setMessage(final Frame frame, final Label label, String text,
                                   boolean success, final int millis) {

        if (success) label.setForeground(Color.green.darker());
        else label.setForeground(Color.red);

        final int chars = 60;
        while (text.length() < chars) text += " ";
        label.setText(text);

        if (!System.getProperty("os.name", "").equals("Mac OS X")) {
            frame.pack();  // Necessary for label not be be cut-off in the middle of the text (layout)
        }

        new Thread(new Runnable() {
            public void run() {
                try {
                    if (millis >= 0) Thread.sleep(millis);
                    String text = "";
                    while (text.length() < chars) text += " ";
                    label.setText(text);
                } catch (Exception e) {
                }
            }
        }, "[javainfo] Message Setter").start();
    }

    /**
     * Creates a normalized representation for the properties.
     * <p/>
     *
     * @param props The container with properties
     * @return Binary data, <code>null</code> on error
     * @since 1.6
     */
    public static byte[] normalizeProps(Properties props) {

        String key, line;
        byte[] binaryline;
        ByteArrayOutputStream bstream = new ByteArrayOutputStream();
        Collator collator = Collator.getInstance(Locale.US);
        collator.setStrength(Collator.PRIMARY);

        try {
            // Assemble the binary data from normalized properties. The lines in the format
            // "key=value" are sorted alphabetically with a Vector and calculated. We use
            // a fixed encoding for the conversion from char (String) to byte. The sorting
            // must be done on-the-fly since Collections.sort() isn't available for Java 1.4

            Vector sorter = JavaInfoUtils.sort(props);

            int size = props.size();
            for (int i = 0; i < size; i++) {
                key = ((String) sorter.elementAt(i)).toLowerCase();  // :-)
                line = key + '=' + props.getProperty(key, "").toUpperCase() + '\n';
                binaryline = line.getBytes(JavaInfoUtils.ENCODING);
                bstream.write(binaryline);
            }
            bstream.flush();

            return bstream.toByteArray();
        } catch (Exception e) {
        }
        return null;
    }

    /**
     * Prints the usage message via <code>System.out</code>.
     * <p/>
     *
     * @since 1.6
     */
    public static void printUsage() {

        String name = JavaInfo.class.getName();  // getSimpleName() is still being invented
        name = name.substring(name.lastIndexOf('.') + 1).toLowerCase();

        String sep = System.getProperties().getProperty("line.separator");
        if (sep == null) sep = "\n";

        String msg = getLocalizedText("JavaInfo.usage");
        msg = new String(JavaInfoUtils.replace(msg.toCharArray(), "%NAME%", name));
        msg = new String(JavaInfoUtils.replace(msg.toCharArray(), "\n", sep));
        System.out.println(msg);
        JavaInfo.printCopyright();
    }

    /**
     * Prints the copyright message via <code>System.out</code>.
     * <p/>
     *
     * @since 1.5
     */
    public static void printCopyright() {
        System.out.println(JavaInfoUtils.COPYRIGHT);
    }

    /**
     * Prints the version message via <code>System.out</code>.
     * <p/>
     *
     * @since 1.5
     */
    private static void printVersion() {

        // We recycle the global resource file 'deploy.properties'

        try {
            Properties props = new Properties();
            props.load(JavaInfo.class.getResourceAsStream(JavaInfo.PROPSPATH));

            String sep = System.getProperties().getProperty("line.separator");
            if (sep == null) sep = "\n";

            String version = props.getProperty("VERSION");
            String build = props.getProperty("BUILD");

            String date = props.getProperty("DATE");
            String pattern = props.getProperty("DATE_PATTERN");
            String locale = props.getProperty("DATE_LOCALE");

            SimpleDateFormat formatter = new SimpleDateFormat(pattern, new Locale(locale, ""));
            Date dateobj = formatter.parse(date, new ParsePosition(0));

            // Convert date to ISO format (yyyy-MM-dd)
            formatter = new SimpleDateFormat("yyyy-MM-dd", Locale.US);
            String isodate = formatter.format(dateobj);

            String name = JavaInfo.class.getName();  // getSimpleName() is still being invented
            name = name.substring(name.lastIndexOf('.') + 1);
            String fullversion = name + ' ' + version + " Build " + build + " (" + isodate + ')';

            System.out.println(sep + fullversion + sep);
            JavaInfo.printCopyright();
        } catch (Exception e) {
        }
    }

    /**
     * This can be used for the application be executed standalone.
     * <p/>
     *
     * @param args Array of strings with console arguments
     * @since 1.6
     */
    public static void main(String[] args) {

        boolean headless = false;

        try {
            // Evaluate the command line parameters

            // Step 1: Determine the locale to print the message as expected
            for (int i = 0; i < args.length; i++) {

                if (args[i].equals("-locale")) {

                    // Check out value and set system property accordingly.
                    // We cannot use the ISO 639-2 codes due to Java 1.4 compatibility.
                    if ((i + 1) < args.length) {
                        String locale = args[i + 1];
                        if (locale.equals("en")) Locale.setDefault(Locale.ENGLISH);
                        else if (locale.equals("de")) {
                            Locale.setDefault(Locale.GERMAN);
                            JavaInfoUtils.LF_LOCALE = "ger_DE";  // Web service uses ISO 639-2
                        }
                        // Default is the system default for the user, not "user.language"!
                    }
                }
            }
            if (Locale.getDefault().getLanguage().equals("de")) JavaInfoUtils.LF_LOCALE = "ger_DE";

            // Step 2: Do what the user likes us to
            for (int i = 0; i < args.length; i++) {

                if (args[i].equals("-help") || args[i].equals("-?")) {
                    JavaInfo.printUsage();
                    System.exit(0);
                } else if (args[i].equals("-version")) {
                    JavaInfo.printVersion();
                    System.exit(0);
                } else if (args[i].equals("-stdout")) {

                    // Continue without GUI, write properties to stdout directly
                    byte[] data = JavaInfoUtils.generateData(JAVAINFO_APP_NAME);
                    System.out.write(data);
                    System.exit(0);  // OK
                } else if (args[i].equals("-headless")) headless = true;
            }

            if (headless) JavaInfo.processCLI();
            else JavaInfo.processGUI();
        } catch (Exception e) {
            System.exit(1);  // Error
        }
        if (headless) System.exit(0);  // OK
    }

    /**
     * Gets the localized text from the resource bundle associated with this class.
     * <p/>
     *
     * @param name The name of the variable
     * @return The text, <code>null</code> if the variable does not exist
     * @since 1.6
     */
    public static String getLocalizedText(String name) {

        // Consults the default locale for this
        ResourceBundle bundle = ResourceBundle.getBundle(JavaInfo.class.getName());
        return bundle.getString(name);
    }
}